@@ -9,9 +9,12 @@ from scipy import optimize, spatial
 from .simulator import random_eddy_params, FibrePopulation
 from copy import deepcopy
 import xarray
-from typing import Sequence, Union
+from typing import Sequence, Union, List, Optional, Any
 import sympy as sym
 from functools import lru_cache
+import os.path as op
+import nibabel as nib
+from fsl.data import image
 
 
 def default_dependencies():
@@ -22,14 +25,14 @@ def default_dependencies():
         (e.g., log_g depends on fibre population, but not any of the acquisition parameters, so is only true for depends=fibre)
     """
     all_params = ['theta', 'phi', 'width', 'dphase', 'log_g', 'f_myelin', 'log_g_myelin', 'non_myelin',
-                  'amplitude_se', 'R2_attenuation', 'R2', 'R2i', 'eddy_odd', 'eddy_even']
+                  'amplitude_se', 'R2_attenuation', 'R2', 'R2i', 'eddy_odd', 'eddy_even', 'intercept']
     return xarray.DataArray(
         [
            # fibre, bval, b0_group, t_phase
             [True, False, False, False],  # theta
             [True, False, False, False],  # phi
             [True, True, False, False],   # width
-            [True, True, True, True],    # dphase
+            [True, True, True, True],     # dphase
             [True, True, False, False],   # log_g
             [True, True, False, False],   # f_myelin
             [True, False, False, False],  # log_g_myelin
@@ -40,6 +43,7 @@ def default_dependencies():
             [True, False, True, False],   # R2i
             [False, True, True, True],    # eddy_odd
             [False, True, True, False],   # eddy_even
+            [False, False, False, False], # intercept
         ],
         coords=[
             ('param', all_params),
@@ -122,6 +126,7 @@ class Parameters:
             ('R2i', 0.002),
             ('eddy_odd', np.zeros(neddy_odd)),
             ('eddy_even', np.zeros(neddy_even)),
+            ('intercept', 0),
         ]:
             depends_on = dependencies.sel(param=name).values
             shell_indices = []
@@ -141,6 +146,42 @@ class Parameters:
 
         return cls(as_dict, shells, eddy_lmax)
 
+    @classmethod
+    def from_bedpostx(cls, shells: Union[MultiShell, Sequence[Shell]], bedpost_dir, voxel, eddy_lmax=1, dependencies=None, pv_thresh=0.05):
+        """Initialises dyad number and orientations based on bedpostX directory
+
+        :param bedpost_dir: bedpostX directory
+        :param voxel: tuple with the (i, j, k) of the voxel
+        :param eddy_lmax: maximum order of the spherical harmonics used to fit the eddy currents
+        :param dependencies: on what observatinal variables does each parameter depend
+            (defaults to result of :func:`default_dependencies`)
+        :param pv_thesh: partial volume threshold to apply to select valid dyads
+        """
+        try:
+            image.addExt(op.join(bedpost_dir, 'dyads1'), mustExist=True)
+        except image.PathError:
+            raise IOError(f"{bedpost_dir} is not a valid bedpostX directory; could not find dyads1")
+
+        angles = []
+        for dyad_index in (1, 2, 3):
+            try:
+                dyad_fn = image.addExt(op.join(bedpost_dir, f'dyads{dyad_index}'), mustExist=True)
+            except image.PathError:
+                break
+            pv_fn = image.addExt(op.join(bedpost_dir, f'mean_f{dyad_index}samples'), mustExist=True)
+            dyad = nib.load(dyad_fn).dataobj[voxel]
+            pv = nib.load(pv_fn).dataobj[voxel]
+            if pv < pv_thresh:
+                break
+            angles.append(spherical.cart2spherical(*dyad)[1:])
+        
+        res = cls.empty(shells, nfibres=len(angles), eddy_lmax=eddy_lmax, dependencies=dependencies)
+        for idx_angle, angle in enumerate(("phi", "theta")):
+            for fibre_list in res.as_dict[angle][1]:
+                for idx_fibre in range(res.nfibres):
+                    fibre_list[idx_fibre] = np.array(angles[idx_fibre][idx_angle])
+        return res
+
     def guess_orient(self, all_data):
         """
         Guesses the fibre orientation fits and updates them in place
@@ -311,6 +352,11 @@ class Parameters:
             self.eddy_lmax
         )
 
+    def simplify_params(self, ):
+        for _, params in self:
+            for p in params:
+                p['dphase'] = ((p['dphase'] + np.pi) % (2 * np.pi)) - np.pi
+
     def to_xarray(self, ):
         """
         Converts all parameters (excluding movement_phase) to an xarray object for later analysis
@@ -323,7 +369,7 @@ class Parameters:
                     as_dict[name].append(vals)
         second_dim = {'eddy_odd': ('m_odd', ), 'eddy_even': ('m_even', )}
         res = {name: xarray.DataArray(
-            arr, coords={'shell': self.shells.coords1d},
+            arr, coords={'shell': self.shells.coords1d, 'fibre': np.arange(1, self.nfibres + 1)},
             dims=('shell', 'fibre') + second_dim.get(name, ())) for name, arr in as_dict.items()}
         return xarray.Dataset(res, attrs={'eddy_lmax': self.eddy_lmax})
 
@@ -388,7 +434,7 @@ class Model(object):
     """
     Model to fit to the DIPPI data
     """
-    def __init__(self, shells: MultiShell, T2=False, dual_phase=False, only_phase=False):
+    def __init__(self, shells: Sequence[Shell], T2=False, dual_phase=False, only_phase=False, with_intercept=False):
         """
         Creates a new DIPPI model
 
@@ -396,6 +442,7 @@ class Model(object):
         :param T2: if True fit R2 and R2* instead of fitting each amplitude independently
         :param dual_phase: if True fit unmyelinated and myelinated axons as separate compartments
         :param only_phase: Ignore the biophysical model and just fit the phase difference between readouts for every dyad
+        :param with_intercept: Adds a phase offset at t_phase=0 to the DIPPI model
         """
         self.shells = shells
         self.T2 = T2
@@ -403,12 +450,13 @@ class Model(object):
             raise ValueError("dual_phase and only_phase are conflicting model definitions")
         self.dual_phase = dual_phase
         self.only_phase = only_phase
+        self.with_intercept = with_intercept
 
         amplitude_se, R2_attenuation, R2, R2i, width, theta, phi = sym.symbols(
             'amplitude_se, R2_attenuation, R2, R2i, width, theta, phi'
         )
-        movement_phase, eddy_phase, non_myelin, log_g, log_g_myelin, f_myelin, dphase = sym.symbols(
-            'movement_phase, eddy_phase, non_myelin, log_g, log_g_myelin, f_myelin, dphase'
+        intercept, movement_phase, eddy_phase, non_myelin, log_g, log_g_myelin, f_myelin, dphase = sym.symbols(
+            'intercept, movement_phase, eddy_phase, non_myelin, log_g, log_g_myelin, f_myelin, dphase'
         )
         bvec = sym.symbols('bvec:3')
 
@@ -425,6 +473,8 @@ class Model(object):
             signal_se = magnitude_se * sym.exp(1j * movement_phase)
 
             confound_phase = movement_phase + (non_myelin * shell.t_phase) + eddy_phase
+            if with_intercept:
+                confound_phase = confound_phase + intercept
             if dual_phase:
                 angle = myelin_phase(shell, phi, theta, log_g_myelin)
                 norm_myelin = sym.sqrt(((1 - f_myelin) + f_myelin * sym.cos(angle)) ** 2 + (f_myelin * sym.sin(angle)) ** 2)
@@ -474,7 +524,7 @@ class Model(object):
         :param to_vary: names of the parameters that should not be filled in yet (None placeholder)
         :return: for each shell a list of arguments (which are the same for the first and second readout)
         """
-        res = []
+        res: List[List[List[Optional[np.ndarray]]]] = []
         for shell, sparams in parameters:
             res.append([])
             for param in sparams:
@@ -487,7 +537,7 @@ class Model(object):
                         eddy_phase += shell.spherical_harmonics(parameters.eddy_lmax, odd=False).dot(param['eddy_even'])
                 for name in self.parameter_names:
                     if name == 'eddy_phase':
-                        value = eddy_phase
+                        value: Optional[np.ndarray] = eddy_phase
                     elif name.startswith('bvec'):
                         idx = int(name[-1])
                         value = shell.gradients[:, idx]
@@ -508,7 +558,7 @@ class Model(object):
         :param split_fibres: if True return the signal for each fibre population individually
         :return: a list with for each shell the signal during the first and second readout
         """
-        res = []
+        res: List[List[Any]] = []
         for shell, all_params, all_funcs in zip(parameters.shells, self.func_args(parameters), self.funcs(only_mag=only_mag)):
             res.append([])
             for readout_func in all_funcs:
@@ -822,7 +872,7 @@ def get_bounds(parameters: Parameters, tobound):
     for name in tobound:
         if name in bounds_vals:
             use_bounds = True
-        nvals = np.sum([arr.size for arr in parameters.to_change([name])[0]])
+        nvals = int(np.sum([arr.size for arr in parameters.to_change([name])[0]]))
 
         bounds.extend([bounds_vals.get(name, (-np.inf, np.inf))] * nvals)
     if not use_bounds:
@@ -831,7 +881,7 @@ def get_bounds(parameters: Parameters, tobound):
 
 
 def fit(parameters, data, model: Model, even_eddy=False, only_myelin=False, movement_phase=True,
-        nfit=100, return_all=False, fit_2D=False, verbose=False) -> Parameters:
+        nfit=100, fit_2D=False, verbose=False) -> Parameters:
     """
     Multi-step fitting procedure for diffusion-weighted phase data in a single voxel
 
@@ -842,7 +892,6 @@ def fit(parameters, data, model: Model, even_eddy=False, only_myelin=False, move
     :param only_myelin: don't fit the non-myelin contribution to the phase
     :param movement_phase: include random phase offsets before the first readout
     :param nfit: maximum number of fits during global optimisation
-    :param return_all: return all local fits of global optimisation rather than just be best one
     :param fit_2D: fits data with gradients only in the x-y plane
     :param verbose: if True report on intermediate fits (2 for very verbose)
     :return: best-fit parameters
@@ -854,16 +903,16 @@ def fit(parameters, data, model: Model, even_eddy=False, only_myelin=False, move
             for fibre_params, phi in zip(shell_params, np.linspace(0, np.pi, parameters.nfibres + 1)):
                 fibre_params['theta'] = np.pi/2.
                 fibre_params['phi'] = phi
-        angles = ('phi', )
+        angles = ['phi', ]
     else:
         #parameters.guess_orient(data)
-        angles = ('theta', 'phi')
+        angles = ['theta', 'phi']
     #print("orientation after DTI:", parameters.as_dict['theta'], parameters.as_dict['phi'])
-    ase_mags = ('R2', 'R2i') if model.T2 else ('R2_attenuation', )
-    best_fit = single_fit(parameters, data, ('amplitude_se', ), model, only_mag=True, verbose=verbose)
+    ase_mags = ['R2', 'R2i'] if model.T2 else ['R2_attenuation']
+    best_fit = single_fit(parameters, data, ['amplitude_se'], model, only_mag=True, verbose=verbose)
     best_fit = single_fit(best_fit, data, ase_mags, model, only_mag=True, verbose=verbose)
-    best_fit = single_fit(best_fit, data, ('width', 'amplitude_se') + ase_mags, model, only_mag=True, verbose=verbose)
-    best_fit = single_fit(best_fit, data, angles + ('width', 'amplitude_se') + ase_mags, model,
+    best_fit = single_fit(best_fit, data, ['width', 'amplitude_se'] + ase_mags, model, only_mag=True, verbose=verbose)
+    best_fit = single_fit(best_fit, data, angles + ['width', 'amplitude_se'] + ase_mags, model,
                           only_mag=True, verbose=verbose)
 
     # fitting phase as well
@@ -877,19 +926,26 @@ def fit(parameters, data, model: Model, even_eddy=False, only_myelin=False, move
         phase_confounds.append('eddy_even')
     if not only_myelin and not model.only_phase:
         phase_confounds.append('non_myelin')
+    if model.with_intercept:
+        phase_confounds.append('intercept')
 
-    log_g_vars = ('log_g_myelin', 'f_myelin') if model.dual_phase else (
-        ("dphase", ) if model.only_phase else ('log_g', ))
+    if model.dual_phase:
+        log_g_vars = ['log_g_myelin', 'f_myelin']
+    elif model.only_phase:
+        log_g_vars = ['dphase', ]
+    else:
+        log_g_vars = ['log_g', ]
     final_fits = []
     best_value = np.inf
-    best_iter_fit = None
     nconsistent = 0
     for _ in range(nfit):
         iter_fit = best_fit.copy()
         for _, params in iter_fit:
             if not only_myelin:
                 params[0]['non_myelin'] = np.random.rand() * 4 - 2
-            for idx_fibre, pfibre in enumerate(params):
+            if model.with_intercept:
+                params[0]['intercept'] = np.random.rand() * 2 * np.pi
+            for _, pfibre in enumerate(params):
                 if model.dual_phase:
                     pfibre['log_g_myelin'] = np.log(np.random.rand() * 0.2 + 0.6)
                     pfibre['f_myelin'] = np.random.rand()
@@ -897,7 +953,6 @@ def fit(parameters, data, model: Model, even_eddy=False, only_myelin=False, move
                     pfibre['dphase'] = (np.random.rand() * 2 - 1) * np.pi
                 else:
                     pfibre['log_g'] = np.log(np.random.rand() * 0.4 + 0.6)
-        phase_confounds = tuple(phase_confounds)
         iter_fit = single_fit(iter_fit, data, phase_confounds, model,
                               verbose=verbose)
 
@@ -914,10 +969,9 @@ def fit(parameters, data, model: Model, even_eddy=False, only_myelin=False, move
             nconsistent += 1
         if nconsistent >= 5:
             break
-    if return_all:
-        return final_fits
-    else:
-        mp = ("movement_phase", ) if movement_phase else ()
-        return single_fit(best_iter_fit, data, angles + ('width', ) + mp +
-                          log_g_vars + ase_mags + phase_confounds, model, verbose=verbose,
-                          scipy_kwargs=dict(options=dict(ftol=1e-13), method='L-BFGS-B'))
+    mp = ["movement_phase", ] if movement_phase else []
+    res = single_fit(best_iter_fit, data, angles + ['width', ] + mp +
+                        log_g_vars + ase_mags + phase_confounds, model, verbose=verbose,
+                        scipy_kwargs=dict(options=dict(ftol=1e-13), method='L-BFGS-B'))
+    res.simplify_params()
+    return res
